import io
import unittest

from TwitterArchive.core import OAuthRequestHandler


class HTTPRequestHandlerTestCase(unittest.TestCase):
    def test_handler(self):
        host = "127.0.0.1"
        port = 8080
        token = "/foo/bar/123?foo"
        expected = f"https://{host}:{port}{token}"

        class MockRequest:
            def makefile(self, *args, **kwargs):
                return io.BytesIO(f"GET {token}".encode())

            def sendall(self, payload):
                return payload

        class MockServer:
            def __init__(server_self, ip_port, Handler):
                handler = Handler(MockRequest(), ip_port, server_self)
                self.assertEqual(expected, handler.token)

        MockServer((host, port), OAuthRequestHandler)



"""Main entrypoint to the program from CLI."""
import argparse
import itertools
import json
import logging
import shutil
from multiprocessing.pool import ThreadPool
from pathlib import Path

from . import __version__
from .core import auth, download_tweet, get_bookmarks


class CapitalizedHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    """Use a more sane capitaliztion in help argparse help menus."""

    def __init__(self, prog: str) -> None:
        """Use the terminal width to determine width."""
        width, height = shutil.get_terminal_size((80, None))
        super().__init__(prog, width=width)

    def add_usage(self, usage, actions, groups, prefix=None):
        """Usage formatter."""
        if prefix is None:
            prefix = "Usage: "
        return super(CapitalizedHelpFormatter, self).add_usage(
            usage,
            actions,
            groups,
            prefix,
        )


def nat_int(s: str) -> int:
    """Type validator for natural integers."""
    int_val = int(s)
    if int_val < 1:
        raise argparse.ArgumentTypeError("Cannot be less than 1")
    return int_val


def build_parser(exit_on_error: bool = True) -> argparse.ArgumentParser:
    """Build the CLI parser.

    :param exit_on_error: Terminate the program (sys.exit) if there is a
                          semantic failure with CLI flags.
    """
    parser = argparse.ArgumentParser(
        add_help=False,
        description="A CLI Tool to archive tweets",
        formatter_class=CapitalizedHelpFormatter,
        exit_on_error=exit_on_error,
    )

    # Set some more sane capitalization
    parser._positionals.title = "Positional Arguments"
    parser._optionals.title = "Options"

    parser.add_argument(
        "--client-id",
        action="store",
        metavar="ID",
        help="Specify the client ID.",
    )
    parser.add_argument(
        "--client-secret",
        action="store",
        metavar="ID",
        help="Specify the client ID.",
    )

    parser.add_argument(
        "--headless",
        action="store_true",
        help="Don't use interactive authentication.",
    )
    parser.add_argument(
        "--no-clobber",
        action="store_true",
        help="Don't redownload/overwrite existing media.",
    )
    parser.add_argument(
        "--num-download-threads",
        metavar="N",
        type=nat_int,
        default=8,
        help="Number of threads to use while downloading media.",
    )
    parser.add_argument(
        "--quiet",
        action="store_true",
        help="Disable download progress bars",
    )
    parser.add_argument(
        "-o",
        "--media-output",
        metavar="FILE",
        action="store",
        type=Path,
        default=Path("./media"),
        help="Path to output downloaded media.",
    )

    # Mutually exclusive options.
    manifest_group = parser.add_mutually_exclusive_group()
    manifest_group.add_argument(
        "-i",
        "--manifest-input",
        action="store",
        metavar="FILE",
        type=Path,
        help="Use an existing manifest and download all media.",
    )
    manifest_group.add_argument(
        "-m",
        "--manifest-output",
        action="store",
        metavar="FILE",
        type=Path,
        default=Path("./bookmark-manifest.json"),
        help="Path to output bookmark manifest.",
    )

    # Generic
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    parser.add_argument(
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message ane exit.",
    )

    return parser


def _setup_logging(lvl: int) -> logging.Logger:
    if lvl == 0:
        log_level = logging.ERROR
    elif lvl == 1:
        log_level = logging.WARNING
    elif lvl == 2:
        log_level = logging.INFO
    else:
        log_level = logging.DEBUG

    logging.basicConfig(
        level=logging.ERROR,  # Third party libs use this log level
        format="%(asctime)s %(name)-16s %(levelname)-8s %(threadName)s %(message)s",
    )

    logger = logging.getLogger("twitter-archive")
    # My code uses this log level
    logger.setLevel(log_level)
    logger.debug(
        "Initialized logger with log level %s (verbose=%s)",
        logging._levelToName[log_level],
        lvl,
    )

    return logger


def main() -> None:
    """Entrypoint from CLI.

    Parse the CLI arguments, and download all the tweets.
    """
    parser = build_parser()
    args = parser.parse_args()
    args = vars(args)

    logger = _setup_logging(args["verbose"])

    logger.debug("Authenticating")
    client = auth(
        headless=args["headless"],
        client_id=args["client_id"],
        client_secret=args["client_secret"],
        use_dotenv=True,
    )

    if args["manifest_input"] is not None:
        logger.info("Loaded existing manifest from '%s'", args["manifest_input"])
        with open(args["manifest_input"], "r") as fp:
            tweets = json.load(fp)
    else:
        logger.info("Fetching manifest from Twitter.")
        tweets = get_bookmarks(client, save_path=args["manifest_output"])

    base_dir = Path(args["media_output"])
    base_dir.mkdir(exist_ok=True, parents=True)

    num_download_threads = args["num_download_threads"]
    logger.info("Downloading all media with %d threads", num_download_threads)
    if num_download_threads == 1:
        for tweet in tweets:
            download_tweet(tweet, base_dir, not args["no_clobber"], args["quiet"])
    else:
        pool = ThreadPool(num_download_threads)
        payloads = zip(
            tweets,
            itertools.repeat(base_dir),
            itertools.repeat(not args["no_clobber"]),
            itertools.repeat(args["quiet"]),
        )
        pool.starmap(download_tweet, payloads)



"""Core internals of the application.

Contains the major interactions with external APIs.
Handles authentication with Twitter, and fetching all necessary Tweets.
"""
import datetime
import http.server
import json
import logging
import os
import socketserver
import time
from json import JSONEncoder
from pathlib import Path
from socket import socket
from typing import Any, Optional
from urllib.parse import urlparse, urlunparse

import requests
import tweepy
from tqdm import tqdm

logger = logging.getLogger("twitter-archive.core")


class OneShotTCPServer(socketserver.TCPServer):
    """TCP server to handle a single HTTP request."""

    def serve_forever(self) -> None:
        """Remove unsupported method from parent server.

        :raises: NotImplementedError
        """
        raise NotImplementedError(
            "Use socketserver.TCPServer to handle several requests instead."
        )

    def process_request(self, request: socket, client_address: str) -> Any:
        """Process a single HTTP request to the server.

        :param request: Request from a client to handle.
        :param client_address: Address of the client sending the request.
        :returns: Instance of the request handler.
        """
        handler_inst = self.finish_request(request, client_address)
        self.shutdown_request(request)
        return handler_inst

    def finish_request(self, request: socket, client_address: str) -> Any:
        """Handoff a single request to the handler.

        :param request: Request from a client to handle.
        :param client_address: Address of the client sending the request.
        :returns: Instance of the request handler.
        """
        return self.RequestHandlerClass(request, client_address, self)

    def _handle_request_noblock(self) -> Any:
        try:
            request, client_address = self.get_request()
        except OSError:
            return

        if self.verify_request(request, client_address):
            try:
                # Almost the same as the standard library implementation,
                # but returns an instance of the request handler.
                return self.process_request(request, client_address)
            except Exception:
                self.handle_error(request, client_address)
                self.shutdown_request(request)
        else:
            self.shutdown_request(request)


class OAuthRequestHandler(http.server.BaseHTTPRequestHandler):
    """Handle a OAuth success request."""

    success_resp = r"""<!doctype html>
<html lang="en">
<h1>
Success, you may close this tab!
</h1>
</html>
    """.encode()

    def __init__(self, request: socket, client_address: str, server: str):
        """Initialze a new instance of the handler.

        :param request: Request to handle.
        :param client_address: Address of the client sending the request.
        :param server: Address of the server itself.
        """
        self._token = None
        super().__init__(request, client_address, server)

    def do_GET(self) -> None:
        """Handle a GET request.

        Sends back the OAuth token as a parameter in the URL.
        """
        # Normalize the URL.
        host, port = self.client_address
        self._token = urlunparse(urlparse(f"https://{host}:{port}{self.path}"))

        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

        self.wfile.write(self.success_resp)

    @property
    def token(self) -> str:
        """Retrieve the token.

        :returns: Token assuming successful authentication.
        """
        return self._token


class TweetEncoder(JSONEncoder):
    """Custom encoder to serialize a series of Tweets."""

    def default(self, obj: Any) -> dict:
        """Serialize Tweets, handling recursive cases.

        :param obj: Object to be serialized
        :returns: Ready to be written as JSON, dict equivalent of obj.
        """
        if isinstance(obj, datetime.datetime):
            return str(obj.isoformat())
        elif isinstance(obj, tweepy.ReferencedTweet):
            return obj.data
        else:
            return super().default(obj)


def _oauth(
    save_path: Path,
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    use_dotenv: bool = False,
    headless: bool = False,
) -> dict:
    """Authenticate with the Twitter API using OAuth 2.X.

    The client ID and secret parameters, if not specified, are loaded from the
    'TWITTER_ARCHIVE_CLIENT_ID' and 'TWITTER_ARCHIVE_CLIENT_SECRET' environment
    variables respectively.

    :param save_path: Path to save the access token.
    :param client_id: Client ID from the Twitter dev app portal.
    :param client_secret: Client secret from the Twitter dev app portal.
    :param use_dotenv: Load a .env file automatically.
    :param headless: Disable interactive authentication.

    :returns: Access token details.

    :raises: ValueError - Missing client ID: client ID not specified and
                          'TWITTER_ARCHIVE_CLIENT_ID' is unset.
    :raises: ValueError - Missing client secret: client secret not specified and
                          'TWITTER_ARCHIVE_CLIENT_ID' is unset.
    """
    REDIRECT_PORT = 8080
    REDIRECT_URI = f"http://localhost:{REDIRECT_PORT}"
    if use_dotenv:
        from dotenv import load_dotenv

        load_dotenv()
        logger.debug("Loaded dotenv")

    client_id = client_id or os.environ.get("TWITTER_ARCHIVE_CLIENT_ID")
    client_secret = client_secret or os.environ.get("TWITTER_ARCHIVE_CLIENT_SECRET")

    if client_id is None:
        raise ValueError("Missing client ID")
    if client_secret is None:
        raise ValueError("Missing client secret")

    oauth2_user_handler = tweepy.OAuth2UserHandler(
        client_id=client_id,
        redirect_uri=REDIRECT_URI,
        scope=["tweet.read", "users.read", "bookmark.read"],
        client_secret=client_secret,
    )

    auth_url = oauth2_user_handler.get_authorization_url()
    print(auth_url)
    logger.info("Prompted user with token url: '%s'", auth_url)

    if not headless:
        # Start the HTTP server to receive the token.
        handler = OAuthRequestHandler
        with OneShotTCPServer(("", REDIRECT_PORT), handler) as httpd:
            handler_inst = httpd.handle_request()
            response_url = handler_inst.token
            logger.info("Received token from HTTP server: '%s'", response_url)
    else:
        response_url = input("Enter the response url here: ")
        logger.info("Received token headless: '%s'", response_url)

    access_token = oauth2_user_handler.fetch_token(response_url)
    with open(save_path, "w") as fp:
        json.dump(access_token, fp)
        logger.info("Wrote access token details to '%s'", save_path)

    return access_token


def auth(
    cache_path: str = "access_token.json",
    use_cache: bool = True,
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    use_dotenv: bool = False,
    headless: bool = False,
) -> tweepy.Client:
    """Login to the Twitter API.

    A wrapper of the _oauth() method, but uses a cache to avoid reauthenticating
    if possible.

    The client ID and secret parameters, if not specified, are loaded from the
    'TWITTER_ARCHIVE_CLIENT_ID' and 'TWITTER_ARCHIVE_CLIENT_SECRET' environment
    variables respectively.

    :param cache_path: Path to cache access token details.
    :param use_cache: Whether to attempt loading token details from cache.
    :param client_id: Client ID from the Twitter dev app portal.
    :param client_secret: Client secret from the Twitter dev app portal.
    :param use_dotenv: Load a .env file automatically.
    :param headless: Disable interactive authentication.

    :returns: Instance of the tweepy client.

    :raises: ValueError: Missing client ID, client ID not specified and
                         'TWITTER_ARCHIVE_CLIENT_ID' is unset.
    :raises: ValueError: Missing client secret, client secret not specified and
                         'TWITTER_ARCHIVE_CLIENT_ID' is unset.
    """
    cache_path = None if cache_path is None else Path(cache_path)
    if not use_cache or not cache_path.is_file():
        logger.info("No cached token available. Reauthenticating.")
        access_token = _oauth(
            cache_path,
            client_id=client_id,
            client_secret=client_secret,
            use_dotenv=use_dotenv,
            headless=headless,
        )
    else:
        # Try to use the cache if possible.
        with open(cache_path, "r") as json_file:
            access_token = json.load(json_file)
        logger.info("Loaded cached access token from '%s'", cache_path)

        # Check if expired, if so re-auth.
        if float(access_token["expires_at"]) < time.time():
            logger.info("Cached token is expired. Reauthenticating.")
            access_token = _oauth(
                cache_path,
                client_id=client_id,
                client_secret=client_secret,
                use_dotenv=use_dotenv,
                headless=headless,
            )

    client = tweepy.Client(
        access_token["access_token"],
        wait_on_rate_limit=True,
    )

    return client


def get_bookmarks(client: tweepy.Client, save_path: Optional[Path] = None) -> dict:
    """Fetch all bookmarked tweets.

    :param client: Authenticated Twitter user whos bookmarks to fetch.
    :param save_path: Path to save manifest of all the tweets.

    :returns: Serialized dict of all the tweets.
    """
    media = {}
    tweets = []

    page_token = None
    while True:
        logging.info("Querying twitter api for page (%s) of bookmarks.", page_token)
        resp = client.get_bookmarks(
            expansions=[
                "attachments.poll_ids",
                "attachments.media_keys",
                "author_id",
                "geo.place_id",
                "referenced_tweets.id",
            ],
            max_results=100,
            pagination_token=page_token,
            media_fields=[
                "media_key",
                "type",
                "url",
                "duration_ms",
                "height",
                "width",
                "alt_text",
                "variants",
            ],
            place_fields=[
                "full_name",
                "id",
                "geo",
            ],
            poll_fields=[
                "id",
                "options",
                "duration_minutes",
                "end_datetime",
                "voting_status",
            ],
            tweet_fields=[
                "id",
                "text",
                "attachments",
                "author_id",
                "conversation_id",
                "created_at",
                "geo",
                "lang",
                "possibly_sensitive",
                "referenced_tweets",
                "source",
            ],
            user_fields=["id"],
        )

        try:
            for i in resp.includes["media"]:
                media[i.media_key] = i
        except KeyError:
            pass

        for i in resp.data:
            if i.data.get("attachments", {}).get("media_keys") is None:
                tweets.append(dict(i))
                continue

            tweet = dict(i)
            tweet["media"] = []
            for media_key in i["attachments"]["media_keys"]:
                tweet["media"].append(media[media_key].data)
                logger.info("Found media for tweet, '%s'", media_key)

            tweets.append(tweet)

        # We retrieve tweets in pages of 100.
        # Go until no more exist.
        try:
            page_token = resp.meta["next_token"]
        except KeyError:
            break

    data = json.dumps(tweets, indent=2, cls=TweetEncoder)
    logger.info("Serialized all bookmarked tweets")

    if save_path is not None:
        with open(save_path, "w") as fp:
            logger.info("Wrote manifest to '%s'", save_path)
            fp.write(data)

    return json.loads(data)


def download_tweet(
    tweet_obj: dict,
    base_dir: Path,
    clobber: bool = True,
    disable_progress_bar: bool = True,
    chunk_size: int = 1024,
) -> None:
    """Download media from a single tweet.

    :param tweet_obj: Dict including all the attributes of the tweet.
    :param base_dir: Base directory to save any media.
    :param clobber: Overwrite existing files.
    :param disable_progress_bar: Silence the progress bar.
    :param chunk_size: Chunk size to use while downloading content.
    """
    if "id" not in tweet_obj:
        raise AttributeError("Missing attribute ID, is this a valid tweet object?")

    my_dir = base_dir / str(tweet_obj["id"])
    my_dir.mkdir(exist_ok=True)

    try:
        for media in tweet_obj["media"]:
            type_ = media["type"]
            if type_ == "video":
                # Only save videos with a bitrate
                variants = [x for x in media["variants"] if "bit_rate" in x]
                max_variant = max(variants, key=lambda x: x["bit_rate"])
                url = max_variant["url"]
                logging.debug("Found video\n  URL: '%s'\n  Bitrate: '%s")
            elif type_ == "photo":
                url = media["url"]
                logging.debug("Found photo URL: '%s'")
            else:
                raise NotImplementedError(
                    f"Type '{type_}' not yet supported for downloading"
                )

            dest = my_dir / Path(urlparse(url).path).name
            if dest.exists() and not clobber:
                logger.info("'%s' already exists. Skipping.", dest)
                continue

            resp = requests.get(url, stream=True)
            length = resp.headers.get("content-length")
            with open(dest, "wb") as f:
                logging.info("Downloading to '%s'", dest)
                # No content length header
                if length is None:
                    f.write(resp.content)
                else:
                    # Progress bar
                    length = int(length)
                    num_bars = int(length / chunk_size)

                    for chunk in tqdm(
                        resp.iter_content(chunk_size=chunk_size),
                        ascii=True,
                        disable=disable_progress_bar,
                        total=num_bars,
                        desc=dest.name,
                        leave=True,
                        unit="KB",
                        miniters=1,
                        ncols=80,
                    ):
                        f.write(chunk)
    except KeyError:
        return
    except NotImplementedError as e:
        logger.warning(e)
        return



import argparse
import contextlib
import io
import unittest
from pathlib import Path

from TwitterArchive import __version__
from TwitterArchive.cli import build_parser


def _default_expected_args():
    return {
        "client_id": None,
        "client_secret": None,
        "headless": False,
        "manifest_input": None,
        "manifest_output": Path("bookmark-manifest.json"),
        "media_output": Path("media"),
        "no_clobber": False,
        "quiet": False,
        "num_download_threads": 8,
        "verbose": 0,
    }


class CLITestCase(unittest.TestCase):
    def test_no_args(self):
        parser = build_parser(False)

        argv = []
        args = parser.parse_args(argv)
        args = vars(args)
        self.assertDictEqual(args, _default_expected_args())

    def test_client_id(self):
        parser = build_parser(False)

        argv = ["--client-id", "foobar"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["client_id"] = "foobar"

        self.assertDictEqual(args, expected)

    def test_client_secret(self):
        parser = build_parser(False)

        argv = ["--client-secret", "foobar"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["client_secret"] = "foobar"

        self.assertDictEqual(args, expected)

    def test_headless(self):
        parser = build_parser(False)

        argv = ["--headless"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["headless"] = True

        self.assertDictEqual(args, expected)

    def test_no_clobber(self):
        parser = build_parser(False)

        argv = ["--no-clobber"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["no_clobber"] = True

        self.assertDictEqual(args, expected)

    def test_num_download_threads(self):
        parser = build_parser(False)

        argv = ["--num-download-threads", "4"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["num_download_threads"] = 4

        self.assertDictEqual(args, expected)

    def test_invalid_num_download_threads_raises(self):
        parser = build_parser(False)

        argv = ["--num-download-threads", "-1"]
        self.assertRaises(argparse.ArgumentError, parser.parse_args, argv)

    def test_quiet(self):
        parser = build_parser(False)

        argv = ["--quiet"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["quiet"] = True

        self.assertDictEqual(args, expected)

    def test_media_arg_short(self):
        parser = build_parser(False)

        argv = ["-o", "foobar"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["media_output"] = Path("foobar")

        self.assertDictEqual(args, expected)

    def test_media_arg_long(self):
        parser = build_parser(False)

        argv = ["--media-output", "foobar"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["media_output"] = Path("foobar")

        self.assertDictEqual(args, expected)

    def test_manifest_output_short(self):
        parser = build_parser(False)

        argv = ["-m", "foobar"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["manifest_output"] = Path("foobar")

        self.assertDictEqual(args, expected)

    def test_manifest_output_long(self):
        parser = build_parser(False)

        argv = ["--manifest-output", "foobar"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["manifest_output"] = Path("foobar")

        self.assertDictEqual(args, expected)

    def test_manifest_input_short(self):
        parser = build_parser(False)

        argv = ["-i", "foobar"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["manifest_input"] = Path("foobar")

        self.assertDictEqual(args, expected)

    def test_manifest_input_long(self):
        parser = build_parser(False)

        argv = ["--manifest-input", "foobar"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["manifest_input"] = Path("foobar")

        self.assertDictEqual(args, expected)

    def test_mutual_exclusion(self):
        parser = build_parser(False)

        argv = ["-m", "manifest-output", "-i", "manifest-input"]
        with self.assertRaises(argparse.ArgumentError):
            parser.parse_args(argv)

    def test_many_short(self):
        parser = build_parser(False)

        argv = ["-o", "media-output", "-m", "manifest-output"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["media_output"] = Path("media-output")
        expected["manifest_output"] = Path("manifest-output")

        self.assertDictEqual(args, expected)

    def test_many_long(self):
        parser = build_parser(False)

        argv = [
            "--media-output",
            "media-output",
            "--manifest-output",
            "manifest-output",
        ]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["media_output"] = Path("media-output")
        expected["manifest_output"] = Path("manifest-output")

        self.assertDictEqual(args, expected)

    def test_verbose(self):
        parser = build_parser(False)

        argv = ["-v"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["verbose"] = 1

        self.assertDictEqual(args, expected)

    def test_verbose_multiple_short(self):
        parser = build_parser(False)

        argv = ["-vvv"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["verbose"] = 3

        self.assertDictEqual(args, expected)

    def test_verbose_multiple_long(self):
        parser = build_parser(False)

        argv = ["--verbose", "--verbose", "--verbose"]
        args = parser.parse_args(argv)
        args = vars(args)

        expected = _default_expected_args()
        expected["verbose"] = 3

        self.assertDictEqual(args, expected)

    def test_version(self):
        # Setup
        stdout = io.StringIO()

        parser = build_parser(False)
        argv = ["--version"]
        with contextlib.redirect_stdout(stdout):
            with self.assertRaises(SystemExit):
                parser.parse_args(argv)

        self.assertIn(__version__, stdout.getvalue())

    def test_help(self):
        # Setup
        stdout = io.StringIO()

        parser = build_parser(False)
        argv = ["--help"]
        with contextlib.redirect_stdout(stdout):
            with self.assertRaises(SystemExit):
                parser.parse_args(argv)

        self.assertIn("help", stdout.getvalue())
